@@ -14,6 +14,7 @@ import (
 	"testing"
 
 	"github.com/smallstep/assert"
+	"go.step.sm/crypto/x25519"
 	"golang.org/x/crypto/ssh"
 )
 
@@ -73,6 +74,8 @@ func TestPublicKey(t *testing.T) {
 	ecdsaKey := must(generateECKey("P-256")).(*ecdsa.PrivateKey)
 	rsaKey := must(generateRSAKey(2048)).(*rsa.PrivateKey)
 	ed25519Key := must(generateOKPKey("Ed25519")).(ed25519.PrivateKey)
+	x25519Pub, x25519Priv, err := x25519.GenerateKey(rand.Reader)
+	assert.FatalError(t, err)
 
 	type args struct {
 		priv interface{}
@@ -89,6 +92,8 @@ func TestPublicKey(t *testing.T) {
 		{"rsaPublic", args{&rsaKey.PublicKey}, rsaKey.Public(), false},
 		{"ed25519", args{ed25519Key}, ed25519Key.Public(), false},
 		{"ed25519Public", args{ed25519.PublicKey(ed25519Key[32:])}, ed25519Key.Public(), false},
+		{"x25519", args{x25519Priv}, x25519Pub, false},
+		{"x25519Public", args{x25519Pub}, x25519Pub, false},
 		{"fail", args{[]byte("octkey")}, nil, true},
 	}
 	for _, tt := range tests {
@@ -231,12 +236,21 @@ func TestGenerateKey_rsa(t *testing.T) {
 		wantType reflect.Type
 		wantErr  bool
 	}{
-		{"RSA1024", args{"RSA", "", 1024}, reflect.TypeOf(&rsa.PrivateKey{}), false},
 		{"RSA2048", args{"RSA", "", 2048}, reflect.TypeOf(&rsa.PrivateKey{}), false},
+		{"RSA3072", args{"RSA", "", 3072}, reflect.TypeOf(&rsa.PrivateKey{}), false},
 		{"fail", args{"RSA", "", 1}, nil, true},
+		{"fail size", args{"RSA", "", 1024}, nil, true},
 	}
 	for _, tt := range tests {
 		t.Run(tt.name, func(t *testing.T) {
+			// Get rid of size validation for coverage purposes.
+			if tt.args.size == 1 {
+				tmp := MinRSAKeyBytes
+				MinRSAKeyBytes = 0
+				t.Cleanup(func() {
+					MinRSAKeyBytes = tmp
+				})
+			}
 			got, err := GenerateKey(tt.args.kty, tt.args.crv, tt.args.size)
 			if (err != nil) != tt.wantErr {
 				t.Errorf("GenerateKey() error = %v, wantErr %v", err, tt.wantErr)
@@ -319,9 +333,9 @@ func TestGenerateKeyPair_rsa(t *testing.T) {
 		wantType1 reflect.Type
 		wantErr   bool
 	}{
-		{"RSA1024", args{"RSA", "", 1024}, pubType, privType, false},
 		{"RSA2048", args{"RSA", "", 2048}, pubType, privType, false},
-		{"fail", args{"RSA", "", 1}, nil, nil, true},
+		{"RSA3072", args{"RSA", "", 3072}, pubType, privType, false},
+		{"fail", args{"RSA", "", 1024}, nil, nil, true},
 	}
 	for _, tt := range tests {
 		t.Run(tt.name, func(t *testing.T) {
