@@ -1,6 +1,7 @@
 package pemutil
 
 import (
+	"bytes"
 	"crypto"
 	"crypto/ecdsa"
 	"crypto/ed25519"
@@ -12,7 +13,6 @@ import (
 	"crypto/x509/pkix"
 	"encoding/asn1"
 	"encoding/pem"
-	"io/ioutil"
 	"math/big"
 	"os"
 	"reflect"
@@ -22,6 +22,7 @@ import (
 	"github.com/pkg/errors"
 	"github.com/smallstep/assert"
 	"go.step.sm/crypto/keyutil"
+	"go.step.sm/crypto/x25519"
 	"golang.org/x/crypto/ssh"
 )
 
@@ -34,6 +35,8 @@ const (
 	ed25519PrivateKey
 	rsaPublicKey
 	rsaPrivateKey
+	x25519PublicKey
+	x25519PrivateKey
 )
 
 const (
@@ -123,11 +126,15 @@ var files = map[string]testdata{
 	"testdata/pkcs8/openssl.rsa2048.enc.pem": {rsaPrivateKey, true},
 	"testdata/pkcs8/openssl.rsa4096.pem":     {rsaPrivateKey, false},
 	"testdata/pkcs8/openssl.rsa4096.pub.pem": {rsaPublicKey, false},
+	"testdata/cosign.pub.pem":                {ecdsaPublicKey, false},
+	"testdata/cosign.enc.pem":                {ecdsaPrivateKey, true},
+	"testdata/nebula.pub":                    {x25519PublicKey, false},
+	"testdata/nebula.key":                    {x25519PrivateKey, false},
 }
 
 func readOrParseSSH(fn string) (interface{}, error) {
 	if strings.HasPrefix(fn, "testdata/openssh") && strings.HasSuffix(fn, ".pub.pem") {
-		b, err := ioutil.ReadFile(fn)
+		b, err := os.ReadFile(fn)
 		if err != nil {
 			return nil, err
 		}
@@ -136,7 +143,7 @@ func readOrParseSSH(fn string) (interface{}, error) {
 	return Read(fn)
 }
 
-func TestReadd(t *testing.T) {
+func TestRead(t *testing.T) {
 	var err error
 	var key interface{}
 
@@ -164,6 +171,10 @@ func TestReadd(t *testing.T) {
 				assert.Type(t, &rsa.PublicKey{}, key)
 			case rsaPrivateKey:
 				assert.Type(t, &rsa.PrivateKey{}, key)
+			case x25519PublicKey:
+				assert.Type(t, x25519.PublicKey{}, key)
+			case x25519PrivateKey:
+				assert.Type(t, x25519.PrivateKey{}, key)
 			default:
 				t.Errorf("type %T not supported", key)
 			}
@@ -223,6 +234,103 @@ func TestReadd(t *testing.T) {
 	}
 }
 
+func TestParseCertificate(t *testing.T) {
+	tests := []struct {
+		fn  string
+		err error
+	}{
+		{"testdata/ca.crt", nil},
+		{"testdata/bundle.crt", nil},
+		{"testdata/badca.crt", errors.New("error parsing certificate")},
+		{"testdata/badpem.crt", errors.New("error decoding pem block")},
+		{"testdata/badder.crt", errors.New("error decoding pem block")},
+		{"testdata/openssl.p256.pem", errors.New("error parsing certificate: no certificate found")},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.fn, func(t *testing.T) {
+			b, err := os.ReadFile(tc.fn)
+			if err != nil {
+				t.Fatal(err)
+			}
+			crt, err := ParseCertificate(b)
+			if tc.err != nil {
+				if assert.Error(t, err) {
+					assert.HasPrefix(t, err.Error(), tc.err.Error())
+				}
+			} else {
+				assert.NoError(t, err)
+				assert.Type(t, &x509.Certificate{}, crt)
+			}
+		})
+	}
+}
+
+func TestParseCertificateBundle(t *testing.T) {
+	tests := []struct {
+		fn  string
+		len int
+		err error
+	}{
+		{"testdata/ca.crt", 1, nil},
+		{"testdata/bundle.crt", 2, nil},
+		{"testdata/badca.crt", 0, errors.New("error parsing certificate")},
+		{"testdata/badpem.crt", 0, errors.New("error decoding pem block")},
+		{"testdata/badder.crt", 0, errors.New("error decoding pem block")},
+		{"testdata/openssl.p256.pem", 0, errors.New("error parsing certificate: no certificate found")},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.fn, func(t *testing.T) {
+			b, err := os.ReadFile(tc.fn)
+			if err != nil {
+				t.Fatal(err)
+			}
+			crts, err := ParseCertificateBundle(b)
+			if tc.err != nil {
+				if assert.Error(t, err) {
+					assert.HasPrefix(t, err.Error(), tc.err.Error())
+				}
+			} else {
+				assert.NoError(t, err)
+				assert.Type(t, []*x509.Certificate{}, crts)
+			}
+			assert.Len(t, tc.len, crts)
+		})
+	}
+}
+
+func TestParseCertificateRequest(t *testing.T) {
+	tests := []struct {
+		fn   string
+		opts []Options
+		err  error
+	}{
+		{"testdata/test.csr", nil, nil},
+		{"testdata/badpem.csr", nil, errors.New("error parsing certificate request")},
+		{"testdata/bad.csr", nil, errors.New("error decoding pem block")},
+		{"testdata/ca.crt", nil, errors.New("error parsing certificate request: no certificate found")},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.fn, func(t *testing.T) {
+			b, err := os.ReadFile(tc.fn)
+			if err != nil {
+				t.Fatal(err)
+			}
+			csr, err := ParseCertificateRequest(b)
+			if tc.err != nil {
+				if assert.Error(t, err) {
+					assert.HasPrefix(t, err.Error(), tc.err.Error())
+				}
+			} else {
+				assert.NoError(t, err)
+				assert.Type(t, &x509.CertificateRequest{}, csr)
+			}
+		})
+	}
+}
+
 func TestReadCertificate(t *testing.T) {
 	tests := []struct {
 		fn   string
@@ -232,11 +340,11 @@ func TestReadCertificate(t *testing.T) {
 		{"testdata/ca.crt", nil, nil},
 		{"testdata/ca.der", nil, nil},
 		{"testdata/bundle.crt", []Options{WithFirstBlock()}, nil},
-		{"testdata/bundle.crt", nil, errors.New("error decoding testdata/bundle.crt: contains more than one PEM endoded block")},
+		{"testdata/bundle.crt", nil, errors.New("error decoding testdata/bundle.crt: contains more than one PEM encoded block")},
 		{"testdata/notexists.crt", nil, errors.New("error reading testdata/notexists.crt: no such file or directory")},
 		{"testdata/badca.crt", nil, errors.New("error parsing testdata/badca.crt")},
 		{"testdata/badpem.crt", nil, errors.New("error decoding testdata/badpem.crt: not a valid PEM encoded block")},
-		{"testdata/badder.crt", nil, errors.New("error parsing testdata/badder.crt: asn1: syntax error: data truncated")},
+		{"testdata/badder.crt", nil, errors.New("error parsing testdata/badder.crt")},
 		{"testdata/openssl.p256.pem", nil, errors.New("error decoding PEM: file 'testdata/openssl.p256.pem' does not contain a certificate")},
 	}
 
@@ -267,7 +375,7 @@ func TestReadCertificateBundle(t *testing.T) {
 		{"testdata/notexists.crt", 0, errors.New("error reading testdata/notexists.crt: no such file or directory")},
 		{"testdata/badca.crt", 0, errors.New("error parsing testdata/badca.crt")},
 		{"testdata/badpem.crt", 0, errors.New("error decoding PEM: file 'testdata/badpem.crt' contains unexpected data")},
-		{"testdata/badder.crt", 0, errors.New("error parsing testdata/badder.crt: asn1: syntax error: data truncated")},
+		{"testdata/badder.crt", 0, errors.New("error parsing testdata/badder.crt")},
 		{"testdata/openssl.p256.pem", 0, errors.New("error decoding PEM: file 'testdata/openssl.p256.pem' is not a certificate bundle")},
 	}
 
@@ -287,7 +395,7 @@ func TestReadCertificateBundle(t *testing.T) {
 	}
 }
 
-func TestParsePEM(t *testing.T) {
+func TestParse(t *testing.T) {
 	type ParseTest struct {
 		in      []byte
 		opts    []Options
@@ -296,7 +404,7 @@ func TestParsePEM(t *testing.T) {
 	}
 	tests := map[string]func(t *testing.T) *ParseTest{
 		"success-ecdsa-public-key": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/openssl.p256.pub.pem")
+			b, err := os.ReadFile("testdata/openssl.p256.pub.pem")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -305,7 +413,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"success-rsa-public-key": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/openssl.rsa1024.pub.pem")
+			b, err := os.ReadFile("testdata/openssl.rsa1024.pub.pem")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -314,7 +422,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"success-rsa-private-key": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/openssl.rsa1024.pem")
+			b, err := os.ReadFile("testdata/openssl.rsa1024.pem")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -323,7 +431,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"success-ecdsa-private-key": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/openssl.p256.pem")
+			b, err := os.ReadFile("testdata/openssl.p256.pem")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -332,7 +440,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"success-ed25519-private-key": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/pkcs8/openssl.ed25519.pem")
+			b, err := os.ReadFile("testdata/pkcs8/openssl.ed25519.pem")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -341,7 +449,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"success-ed25519-enc-private-key": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/pkcs8/openssl.ed25519.enc.pem")
+			b, err := os.ReadFile("testdata/pkcs8/openssl.ed25519.enc.pem")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -350,7 +458,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"success-x509-crt": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/ca.crt")
+			b, err := os.ReadFile("testdata/ca.crt")
 			assert.FatalError(t, err)
 			return &ParseTest{
 				in:      b,
@@ -359,7 +467,7 @@ func TestParsePEM(t *testing.T) {
 			}
 		},
 		"fail-options": func(t *testing.T) *ParseTest {
-			b, err := ioutil.ReadFile("testdata/ca.crt")
+			b, err := os.ReadFile("testdata/ca.crt")
 			assert.FatalError(t, err)
 			err = errors.New("an error")
 			return &ParseTest{
@@ -369,6 +477,57 @@ func TestParsePEM(t *testing.T) {
 				err:     err,
 			}
 		},
+		"fail-password": func(t *testing.T) *ParseTest {
+			b, err := os.ReadFile("testdata/openssl.p256.enc.pem")
+			assert.FatalError(t, err)
+			return &ParseTest{
+				in:      b,
+				opts:    []Options{WithPassword([]byte("badpassword"))},
+				cmpType: ecdsa.PrivateKey{},
+				err:     errors.New("error decrypting PEM"),
+			}
+		},
+		"fail-pkcs8-password": func(t *testing.T) *ParseTest {
+			b, err := os.ReadFile("testdata/pkcs8/openssl.ed25519.enc.pem")
+			assert.FatalError(t, err)
+			return &ParseTest{
+				in:      b,
+				opts:    []Options{WithPassword([]byte("badpassword"))},
+				cmpType: ed25519.PrivateKey{},
+				err:     errors.New("error decrypting PEM: x509: decryption password incorrect"),
+			}
+		},
+		"fail-type": func(t *testing.T) *ParseTest {
+			b, err := os.ReadFile("testdata/openssl.p256.pub.pem")
+			assert.FatalError(t, err)
+			b = bytes.ReplaceAll(b, []byte("PUBLIC KEY"), []byte("EC PUBLIC KEY"))
+			return &ParseTest{
+				in:      b,
+				opts:    []Options{},
+				cmpType: nil,
+				err:     errors.New("error decoding PEM: contains an unexpected header 'EC PUBLIC KEY'"),
+			}
+		},
+		"fail-nebula-pub-size": func(t *testing.T) *ParseTest {
+			b, err := os.ReadFile("testdata/badnebula.pub")
+			assert.FatalError(t, err)
+			return &ParseTest{
+				in:      b,
+				opts:    []Options{},
+				cmpType: nil,
+				err:     errors.New("error parsing PEM: key is not 32 bytes"),
+			}
+		},
+		"fail-nebula-key-size": func(t *testing.T) *ParseTest {
+			b, err := os.ReadFile("testdata/badnebula.key")
+			assert.FatalError(t, err)
+			return &ParseTest{
+				in:      b,
+				opts:    []Options{},
+				cmpType: nil,
+				err:     errors.New("error parsing PEM: key is not 32 bytes"),
+			}
+		},
 	}
 	for name, genTestCase := range tests {
 		t.Run(name, func(t *testing.T) {
@@ -388,6 +547,7 @@ func TestParsePEM(t *testing.T) {
 	}
 }
 
+//nolint:staticcheck // required for legacy compatibility
 func TestSerialize(t *testing.T) {
 	tests := map[string]struct {
 		in    func() (interface{}, error)
@@ -404,12 +564,12 @@ func TestSerialize(t *testing.T) {
 		},
 		"RSA Private Key success": {
 			in: func() (interface{}, error) {
-				return keyutil.GenerateKey("RSA", "", 1024)
+				return keyutil.GenerateKey("RSA", "", 2048)
 			},
 		},
 		"RSA Public Key success": {
 			in: func() (interface{}, error) {
-				pub, _, err := keyutil.GenerateKeyPair("RSA", "", 1024)
+				pub, _, err := keyutil.GenerateKeyPair("RSA", "", 2048)
 				return pub, err
 			},
 		},
@@ -495,7 +655,9 @@ func TestSerialize(t *testing.T) {
 		} else if test.pass != "" && test.file != "" {
 			p, err = Serialize(in, WithPassword([]byte(test.pass)), ToFile(test.file, 0600))
 		} else if test.pass != "" && test.pkcs8 {
-			p, err = Serialize(in, WithPassword([]byte(test.pass)), WithPKCS8(true))
+			p, err = Serialize(in, WithPKCS8(true), WithPasswordPrompt("Please enter the password to encrypt the key", func(prompt string) ([]byte, error) {
+				return []byte(test.pass), nil
+			}))
 		} else if test.pass != "" {
 			p, err = Serialize(in, WithPassword([]byte(test.pass)))
 		} else {
@@ -543,10 +705,6 @@ func TestSerialize(t *testing.T) {
 						assert.Equals(t, p.Type, "ENCRYPTED PRIVATE KEY")
 						actualBytes, err = DecryptPKCS8PrivateKey(p.Bytes, []byte(test.pass))
 						assert.FatalError(t, err)
-						// remove padding
-						length := len(actualBytes)
-						paddginLength := int(actualBytes[length-1])
-						actualBytes = actualBytes[:length-paddginLength]
 					} else {
 						assert.Equals(t, p.Type, "EC PRIVATE KEY")
 						assert.True(t, x509.IsEncryptedPEMBlock(p))
@@ -571,7 +729,7 @@ func TestSerialize(t *testing.T) {
 						assert.Equals(t, fileInfo.Mode(), os.FileMode(0600))
 						// Verify that key written to file is correct
 						var keyFileBytes []byte
-						keyFileBytes, err = ioutil.ReadFile(test.file)
+						keyFileBytes, err = os.ReadFile(test.file)
 						assert.FatalError(t, err)
 						pemKey, _ := pem.Decode(keyFileBytes)
 						assert.Equals(t, pemKey.Type, "EC PRIVATE KEY")
@@ -645,9 +803,9 @@ func TestParseDER(t *testing.T) {
 	ecdsaKey := k2.(*ecdsa.PrivateKey)
 	edKey := k3.(ed25519.PrivateKey)
 	// Ed25519 der files
-	edPubDer, err := ioutil.ReadFile("testdata/pkcs8/openssl.ed25519.pub.der")
+	edPubDer, err := os.ReadFile("testdata/pkcs8/openssl.ed25519.pub.der")
 	assert.FatalError(t, err)
-	edPrivDer, err := ioutil.ReadFile("testdata/pkcs8/openssl.ed25519.der")
+	edPrivDer, err := os.ReadFile("testdata/pkcs8/openssl.ed25519.der")
 	assert.FatalError(t, err)
 
 	toDER := func(k interface{}) []byte {
@@ -708,7 +866,7 @@ func TestParseKey(t *testing.T) {
 			continue
 		}
 		t.Run(fn, func(t *testing.T) {
-			data, err := ioutil.ReadFile(fn)
+			data, err := os.ReadFile(fn)
 			assert.FatalError(t, err)
 			if td.encrypted {
 				key, err = ParseKey(data, WithPassword([]byte("mypassword")))
@@ -731,6 +889,10 @@ func TestParseKey(t *testing.T) {
 				assert.Type(t, &rsa.PublicKey{}, key)
 			case rsaPrivateKey:
 				assert.Type(t, &rsa.PrivateKey{}, key)
+			case x25519PublicKey:
+				assert.Type(t, x25519.PublicKey{}, key)
+			case x25519PrivateKey:
+				assert.Type(t, x25519.PrivateKey{}, key)
 			default:
 				t.Errorf("type %T not supported", key)
 			}
@@ -768,7 +930,7 @@ func TestParseSSH(t *testing.T) {
 			continue
 		}
 		t.Run(fn, func(t *testing.T) {
-			data, err := ioutil.ReadFile(fn)
+			data, err := os.ReadFile(fn)
 			assert.FatalError(t, err)
 			key, err = ParseSSH(data)
 			assert.FatalError(t, err)
@@ -793,6 +955,10 @@ func TestOpenSSH(t *testing.T) {
 		if strings.HasSuffix(fn, ".pub.pem") {
 			continue
 		}
+		// skip x25519 keys
+		if td.typ == x25519PublicKey || td.typ == x25519PrivateKey {
+			continue
+		}
 		// To be able to run this in parallel we need to declare local
 		// variables.
 		fn, td := fn, td
@@ -833,7 +999,7 @@ func TestOpenSSH(t *testing.T) {
 
 func TestRead_options(t *testing.T) {
 	mustKey := func(filename string) interface{} {
-		b, err := ioutil.ReadFile(filename)
+		b, err := os.ReadFile(filename)
 		assert.FatalError(t, err)
 		key, err := ssh.ParseRawPrivateKey(b)
 		assert.FatalError(t, err)
@@ -880,7 +1046,7 @@ func TestRead_options(t *testing.T) {
 
 func TestRead_promptPassword(t *testing.T) {
 	mustKey := func(filename string) interface{} {
-		b, err := ioutil.ReadFile(filename)
+		b, err := os.ReadFile(filename)
 		assert.FatalError(t, err)
 		key, err := ssh.ParseRawPrivateKey(b)
 		assert.FatalError(t, err)
